// --------------------------------------------------
// Multiobjective
// --------------------------------------------------

#include <ctime>
#include <fstream>
#include <iostream>
#include <list>
#include <map>
#include <vector>
#include <ilcplex/ilocplex.h>

#include "bdd.hpp"
#include "bdd_util.hpp"
#include "pareto_util.hpp"
#include "stats.hpp"

#include "knapsack/knapsack_instance.hpp"
#include "knapsack/knapsack_bdd.hpp"

#include "indepset/indepset_instance.hpp"
#include "indepset/indepset_bdd.hpp"

#include "setpacking/setpacking_instance.hpp"

#include "setcovering/setcovering_instance.hpp"
#include "setcovering/setcovering_bdd.hpp"


using namespace std;



//
// Main function
//
int main(int argc, char* argv[]) {
	if (argc != 4) {
		cout << '\n';
		cout << "Usage: multiobj [input file] [problem type] [preprocess]\n";
		
		cout << "\n\twhere:\n";
		cout << "\t\tproblem_type = 1: knapsack\n";
 		cout << "\t\tproblem_type = 2: independent set\n";
 		cout << "\t\tproblem_type = 3: set packing\n";
 		cout << "\t\tproblem_type = 4: set covering\n";
				
		cout << "\n";
		cout << "\t\tpreprocess = 0: no preprocessing\n";
		cout << "\t\tpreprocess = 1: minimize bandwidth (only setpack/setcover)\n";
 		
		cout << "\n";
		exit(1);
	}

	// read input 
	int problem_type = atoi(argv[2]);
	bool preprocess = (atoi(argv[3]) == 1);


	// Initialize timers (for stats) 
	Stats timers;
	int bdd_compilation_time = timers.register_name("BDD compilation time");
	int bdd_reduction_time = timers.register_name("BDD reduction time");
	int bdd_pareto_time = timers.register_name("Pareto time");

	// BDD and objective function coefficients
	BDD* bdd = NULL;
	vector< vector<int> > obj_coefficients;

	// Read problem and construct BDDs
	timers.start_timer(bdd_compilation_time);
  	
	// ---- Knapsack ----
	if (problem_type == 1) {

		// read instance  
		MultiObjKnapsackInstance inst(argv[1]);

		// create initial BDD
		KnapsackBDDConstructor bddConstructor(
			new KnapsackInstance(inst.n_vars, inst.obj_coeffs[0], inst.coeffs, inst.rhs),			 
 			0);

		bdd = bddConstructor.generate_exact();

		// get objective function coefficients
 		obj_coefficients = inst.obj_coeffs;
	}
	
	// ---- Independent Set ---- 
	else if (problem_type == 2) {

		// read instance
 		IndepSetInst inst(argv[1]);
		
		// create artificial objective vector with one unit      
 		vector< vector<int> > obj_coeffs(1, vector<int>(inst.graph->n_vertices, 1));

 		// create BDD with unitary objective function
 		IndepSetBDDConstructor bddConstructor(&inst, obj_coeffs, 0);
 		bdd = bddConstructor.generate_exact();
	}

	// ---- Set Packing ----
	else if (problem_type == 3) {

		// read instance 
 		SetPackingInstance setpack(argv[1]);

		// preprocess
		if (preprocess) {
			setpack.minimize_bandwidth();
		}

		// reduce set packing instance to an independent set problem 
		IndepSetInst* inst = setpack.create_indepset_instance();

		// create artificial objective vector with one unit      
 		vector< vector<int> > obj_coeffs = setpack.objs;

		// create BDD with unitary objective function
 		IndepSetBDDConstructor bddConstructor(inst, obj_coeffs, 0);
 		bdd = bddConstructor.generate_exact();

		// remap objective function vector to consider variable ordering 
		vector<int>& var_layer = bddConstructor.var_layer;
		obj_coefficients.resize(setpack.objs.size(), vector<int>(setpack.n_vars, -1));
		for (size_t p = 0; p < setpack.objs.size(); ++p) {
			for (int i = 0; i < setpack.n_vars; ++i) {
				obj_coefficients[p][i] = setpack.objs[p][ var_layer[i] ];
			}
		}
		
		delete inst;
	}

	// ---- Set Covering ----
	else if (problem_type == 4) {

		// read instance 
		SetCoveringInstance setcover(argv[1]);
		
		// preprocess
		if (preprocess) {
			setcover.minimize_bandwidth();
		}

		// create initial BDD
		SetCoveringBDDConstructor bddConstructor(&setcover, setcover.objs, 0);
 		bdd = bddConstructor.generate_exact();

		// remap objective function vector to consider variable ordering 
 		vector<int>& var_layer = bddConstructor.var_layer;
 		obj_coefficients.resize(setcover.objs.size(), vector<int>(setcover.n_vars, -1));
 		for (size_t p = 0; p < setcover.objs.size(); ++p) {
 			for (int i = 0; i < setcover.n_vars; ++i) {
 				obj_coefficients[p][i] = setcover.objs[p][ var_layer[i] ];
 			}
 		}
	}

	else {
		cout << "\nError: problem type not known.\n";
		exit(1);
	}
	timers.end_timer(bdd_compilation_time);


	// Reduce BDD 
	cout << "\nReducing BDD...\n";
	timers.start_timer(bdd_reduction_time);	
	
	int initial_width = bdd->get_width();
	int initial_node_size = bdd->get_num_nodes();

 	BDDAlg::reduce(bdd);

 	int reduced_width = bdd->get_width();
 	int reduced_node_size = bdd->get_num_nodes();

	cout << "\tinitial width = " << initial_width << '\n';
	cout << "\treduced width = " << reduced_width << '\n'; 

	timers.end_timer(bdd_reduction_time);


	// Generate pareto set 
	timers.start_timer(bdd_pareto_time);
 	
 	ParetoSet* paretoSet = BDDAlg::pareto_set(bdd, obj_coefficients);
 	
	timers.end_timer(bdd_pareto_time);


	// print statistics 
	cout << "Stats: " << endl;
	cout << "\tInitial BDD size: width = " << initial_width << " - nodes = " << initial_node_size << '\n';
	cout << "\tReduced BDD size: width = " << reduced_width << " - nodes = " << reduced_node_size << '\n';
	cout << "\tSize of pareto set: " << paretoSet->sols.size() << '\n';
	cout << "\tBDD compilation time: " << timers.get_time(bdd_compilation_time) << "s\n";
	cout << "\tBDD reduction time: " << timers.get_time(bdd_reduction_time) << "s\n";
	cout << "\tPareto time: " << timers.get_time(bdd_pareto_time) << "s\n";
	cout << '\n';


	return 0;
}
